import os
import pdb
import json

from .get_response import GetResponse


class ClaimVerifier():
    def __init__(self, model_name, label_n=2, cache_dir="./data/cache/", demon_dir="data/demos/",
                 use_external_model=False):
        self.model = None
        self.model_name = model_name
        self.label_n = label_n
        if os.path.isdir(model_name) or use_external_model:
            from unsloth import FastLanguageModel

            self.model, self.tokenizer = FastLanguageModel.from_pretrained(
                model_name=model_name,
                max_seq_length=2048,
                dtype=None,
                load_in_4bit=True,
            )
            FastLanguageModel.for_inference(self.model)
            self.tokenizer.padding_side = "left"
            self.alpaca_prompt = open("./prompt/verification_alpaca_template.txt", "r").read()
            self.instruction = open("./prompt/verification_instruction_binary_no_demo.txt", "r").read()
        else:
            cache_dir = os.path.join(cache_dir, model_name)
            os.makedirs(cache_dir, exist_ok=True)
            self.cache_file = os.path.join(cache_dir, "claim_verification_cache.json")
            self.demon_path = os.path.join(demon_dir, 'few_shot_examples.jsonl')
            self.get_model_response = GetResponse(cache_file=self.cache_file,
                                                  model_name=model_name,
                                                  max_tokens=1000,
                                                  temperature=0)
            self.system_message = "You are a helpful assistant who can verify the truthfulness of a claim against reliable external world knowledge."
            self.prompt_initial_temp = self.get_initial_prompt_template()

    def get_instruction_template(self):
        prompt_temp = ''
        if self.label_n == 2 and "claude" in self.model_name:
            prompt_temp = open("./prompt/verification_instruction_claude_binary.txt", "r").read()
        elif self.label_n == 3 and "claude" in self.model_name:
            prompt_temp = open("./prompt/verification_instruction_claude_trinary.txt", "r").read()
        elif self.label_n == 2 and "claude" not in self.model_name:
            prompt_temp = open("./prompt/verification_instruction_binary.txt", "r").read()
        elif self.label_n == 3 and "claude" not in self.model_name:
            prompt_temp = open("./prompt/verification_instruction_trinary.txt", "r").read()
        return prompt_temp

    def get_initial_prompt_template(self):
        prompt_temp = self.get_instruction_template()
        with open(self.demon_path, "r") as f:
            example_data = [json.loads(line) for line in f if line.strip()]
        element_lst = []
        for dict_item in example_data:
            claim = dict_item["claim"]
            search_result_str = dict_item["search_result"]
            human_label = dict_item["human_label"]
            if self.label_n == 2:
                if human_label == "support":
                    human_label = "Supported."
                else:
                    human_label = "Unsupported."
            if "claude" in self.model_name:
                element_lst.extend([search_result_str, claim, human_label])
            else:
                element_lst.extend([claim, search_result_str, human_label])

        prompt_few_shot = prompt_temp.format(*element_lst)

        if "claude" in self.model_name:
            self.your_task = "Your task:\n\n{search_results}\n\nClaim: {claim}\n\nTask: Given the search results above, is the claim supported or unsupported? Mark your decision with ### signs.\n\nYour decision:"
        else:
            self.your_task = "Your task:\n\nClaim: {claim}\n\n{search_results}\n\nYour decision:"

        """
        choose system message
        """
        if self.label_n == 2:
            self.system_message = "You are a helpful assistant who can judge whether a claim is supported by the search results or not."
        elif self.label_n == 3:
            self.system_message = "You are a helpful assistant who can judge whether a claim is supported or contradicted by the search results, or whether there is no enough information to make a judgement."

        return prompt_few_shot

    def verifying_claim(self, claim_snippets_dict, search_res_num=5, cost_estimate_only=False):
        """
        search_snippet_lst = [{"title": title, "snippet": snippet, "link": link}, ...]
        """
        prompt_tok_cnt, response_tok_cnt = 0, 0
        out_lst = []
        claim_verify_res_dict = {}
        for claim, search_snippet_lst in claim_snippets_dict.items():
            search_res_str = ""
            search_cnt = 1
            for search_dict in search_snippet_lst[:search_res_num]:
                search_res_str += f'Search result {search_cnt}\nTitle: {search_dict["title"].strip()}\nLink: {search_dict["link"].strip()}\nContent: {search_dict["snippet"].strip()}\n\n'
                search_cnt += 1
            if self.model:
                usr_input = f"Claim: {claim.strip()}\n\n{search_res_str.strip()}"
                formatted_input = self.alpaca_prompt.format(self.instruction, usr_input)

                inputs = self.tokenizer(formatted_input, return_tensors="pt").to("cuda")
                output = self.model.generate(**inputs,
                                             max_new_tokens=500,
                                             use_cache=True,
                                             eos_token_id=[self.tokenizer.eos_token_id,
                                                           self.tokenizer.convert_tokens_to_ids("<|eot_id|>")],
                                             pad_token_id=self.tokenizer.eos_token_id, )
                response = self.tokenizer.batch_decode(output)
                clean_output = ' '.join(response).split("<|end_header_id|>\n\n")[
                    -1].replace("<|eot_id|>", "").strip()

            else:
                prompt_tail = self.your_task.format(
                    claim=claim,
                    search_results=search_res_str.strip(),
                )
                prompt = f"{self.prompt_initial_temp}\n\n{prompt_tail}"
                response, prompt_tok_num, response_tok_num = self.get_model_response.get_response(self.system_message,
                                                                                                  prompt,
                                                                                                  cost_estimate_only=cost_estimate_only)
                prompt_tok_cnt += prompt_tok_num
                response_tok_cnt += response_tok_num

                clean_output = response.replace("#", "").split(".")[0].lower() if response is not None else None
            claim_verify_res_dict = {"claim": claim,
                                     "search_results": search_res_str,
                                     "verification_result": clean_output}
            out_lst.append(claim_verify_res_dict)
        return out_lst, prompt_tok_cnt, response_tok_cnt



# get K median and K max
from collections import defaultdict

def get_stats(domain_model_triplet_dict):
    domain_K_dict = defaultdict(lambda: defaultdict(int))
    for domain, model_triplet_dict in domain_model_triplet_dict.items():
        claim_num_lst = []
        for model_name, triplet_lst in model_triplet_dict.items():
            for triplet in triplet_lst:
                claim_num_lst.append(triplet[1])

        claim_num_lst.sort()
        K_median = claim_num_lst[len(claim_num_lst)//2]
        K_max = claim_num_lst[-1]
        domain_K_dict[domain]["K_median"] = K_median
        domain_K_dict[domain]["K_max"] = K_max
        # print(f"{domain} - {K_median}: {K_max}")

    return domain_K_dict

def get_avg_numbers(domain_model_triplet_dict, domain_K_dict):
    for domain, model_triplet_dict in domain_model_triplet_dict.items():
        K_median = domain_K_dict[domain]["K_median"]
        K_max = domain_K_dict[domain]["K_max"]

        table_content = []
        F1_at_median_lst = []
        for model_name in model_triplet_dict.keys():
            triplet_lst = domain_model_triplet_dict[domain][model_name]

            sent_len_lst = [x[2] for x in triplet_lst]
            sup_lst = [x[0] for x in triplet_lst]
            uns_lst = [x[1] - x[0] for x in triplet_lst]
            prec_lst = [x[0] / x[1] for x in triplet_lst]
            rec_med_lst = [min(x[0] / K_median, 1) for x in triplet_lst]
            rec_max_lst = [min(x[0] / K_max, 1) for x in triplet_lst]

            # get f1@K median and f1@K max
            f1_med_lst = [2 * prec * rec_med / (prec + rec_med) if rec_med > 0 else 0 for prec, rec_med in zip(prec_lst, rec_med_lst)]
            f1_max_lst = [2 * prec * rec_max / (prec + rec_max) if rec_max > 0 else 0 for prec, rec_max in zip(prec_lst, rec_max_lst)]

            # get ave. numbers
            ave_sent = sum(sent_len_lst) / len(sent_len_lst)
            S = sum(sup_lst) / len(sup_lst)
            U = sum(uns_lst) / len(uns_lst)
            P = sum(prec_lst) / len(prec_lst)
            Rec_med = sum(rec_med_lst) / len(rec_med_lst)
            Rec_max = sum(rec_max_lst) / len(rec_max_lst)
            F1_med = sum(f1_med_lst) / len(f1_med_lst)
            F1_max = sum(f1_max_lst) / len(f1_max_lst)

            table_row = [model_name, domain, round(ave_sent, 3), round(S, 3), round(U, 3), round(P, 3), round(Rec_med, 3), round(Rec_max, 3), round(F1_med, 3), round(F1_max, 3)]
            table_content.append(table_row)

            F1_at_median_lst.append(100*round(F1_med, 3))

            print(f"[{domain}-{model_name}] \nF1@k median: {F1_med:.3f}, F1@k max: {F1_max:.3f}")

def get_veriscore(domain_model_triplet_dict):
     domain_K_dict= get_stats(domain_model_triplet_dict)
     get_avg_numbers(domain_model_triplet_dict, domain_K_dict)


import os
import json
import argparse

from collections import defaultdict

from tqdm import tqdm

from veriscore import utils
from .claim_verifier import ClaimVerifier

if __name__ == '__main__':
    """
    add argparse for label_n and model_name
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default='./data')
    parser.add_argument("--input_file", type=str, required=True)
    parser.add_argument("--output_dir", type=str, default='./data')
    parser.add_argument("--cache_dir", type=str, default='./data/cache')
    parser.add_argument("--model_name", type=str, default="gpt-4o")
    parser.add_argument("--label_n", type=int, default=3, choices=[2, 3])
    parser.add_argument("--search_res_num", type=int, default=10)
    parser.add_argument("--use_external_model", action='store_true')
    args = parser.parse_args()

    model_name = args.model_name
    label_n = args.label_n

    input_file_name = "".join(args.input_file.split('.')[:-1])
    input_path = os.path.join(args.data_dir, args.input_file)
    with open(input_path, "r") as f:
        input_data = [json.loads(line) for line in f if line.strip()]
    demon_dir = os.path.join(args.data_dir, 'demos')
    claim_verifier = ClaimVerifier(model_name=model_name, label_n=label_n,
                                   cache_dir=args.cache_dir, demon_dir=demon_dir,
                                   use_external_model=args.use_external_model)

    output_dir = os.path.join(args.output_dir, 'model_output')
    os.makedirs(output_dir, exist_ok=True)
    output_file = f'verification_{input_file_name}_{label_n}.jsonl'
    output_path = os.path.join(output_dir, output_file)

    model_domain_triplet_dict = defaultdict(
        lambda: defaultdict(list))  # triplet = [[supported, total, # of sentences], ...]

    total_prompt_tok_cnt = 0
    total_resp_tok_cnt = 0

    with open(output_path, "w") as f:
        for dict_item in tqdm(input_data):
            model_name = dict_item['model']
            domain = dict_item['prompt_source']
            claim_search_results = dict_item["claim_search_results"]

            if dict_item['abstained']:
                f.write(json.dumps(dict_item) + "\n")
                continue

            claim_verify_res_dict, prompt_tok_cnt, response_tok_cnt = claim_verifier.verifying_claim(
                claim_search_results, search_res_num=args.search_res_num)
            dict_item["claim_verification_result"] = claim_verify_res_dict

            f.write(json.dumps(dict_item) + "\n")

            total_prompt_tok_cnt += prompt_tok_cnt
            total_resp_tok_cnt += response_tok_cnt

            ## for VeriScore calculation
            triplet = [0, 0, 0]
            triplet[1] = len(dict_item['all_claims'])
            triplet[2] = len(dict_item['claim_list'])
            if not dict_item['claim_search_results']:
                triplet[0] = 0
            else:
                for claim_veri_res in dict_item['claim_verification_result']:
                    if claim_veri_res['verification_result'] == "supported":
                        triplet[0] += 1
            model_domain_triplet_dict[domain][model_name].append(triplet)


    print(f"claim verification is done! saved to {output_path}")
    utils.get_veriscore(model_domain_triplet_dict)
    print(f"Total cost: {total_prompt_tok_cnt * 10 / 1e6 + total_resp_tok_cnt * 30 / 1e6}")



"""
This script is written to extract claims from the model responses.
"""

import os
import json
import argparse
from tqdm import tqdm
from .claim_extractor import ClaimExtractor

abstain_responses = ["I'm sorry, I cannot fulfill that request.",
                     "I'm sorry, I can't fulfill that request.",
                     "I'm sorry, but I cannot fulfill that request.",
                     "I'm sorry, but I can't fulfill that request.",
                     "Sorry, but I can't fulfill that request.",
                     "Sorry, I can't do that."]

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default='./data')
    parser.add_argument("--input_file", type=str, required=True)
    parser.add_argument("--output_dir", type=str, default='./data')
    parser.add_argument("--cache_dir", type=str, default='./data/cache')
    parser.add_argument("--model_name", type=str, default="gpt-4-0125-preview")
    parser.add_argument("--use_external_model", action='store_true')
    args = parser.parse_args()

    input_file_name = "".join(args.input_file.split('.')[:-1])

    input_path = os.path.join(args.data_dir, args.input_file)
    with open(input_path, "r") as f:
        data = [json.loads(x) for x in f.readlines() if x.strip()]

    # initialize objects
    model_name = args.model_name
    claim_extractor = ClaimExtractor(model_name, args.cache_dir, args.use_external_model)

    output_dir = args.output_dir
    output_file = f"claims_{input_file_name}.jsonl"
    output_path = os.path.join(output_dir, output_file)

    with open(output_path, "w") as f:
        for dict_item in tqdm(data):
            # get necessary info

            response = dict_item["response"]
            prompt_source = dict_item["prompt_source"]
            model = dict_item["model"]

            # skip abstained responses
            if response.strip() in abstain_responses:
                output_dict = {"question": question.strip(),
                               "response": response.strip(),
                               "abstained": True,
                               "prompt_source": prompt_source,
                               "model": model, }
                f.write(json.dumps(output_dict) + "\n")
                continue

            if "question" in dict_item and dict_item["question"]:
                question = dict_item["question"]
                snippet_lst, claim_list, all_claims, prompt_tok_cnt, response_tok_cnt = claim_extractor.qa_scanner_extractor(
                    question, response)
            else:
                question = ''
                snippet_lst, claim_list, all_claims, prompt_tok_cnt, response_tok_cnt = claim_extractor.non_qa_scanner_extractor(
                    response)

            # write output
            output_dict = {"question": question.strip(),
                           "prompt_source": prompt_source,
                           "response": response.strip(),
                           "prompt_tok_cnt": prompt_tok_cnt,
                           "response_tok_cnt": response_tok_cnt,
                           "model": model,
                           "abstained": False,  # "abstained": False, "abstained": True
                           "claim_list": claim_list,
                           "all_claims": all_claims
                           }
            f.write(json.dumps(output_dict) + "\n")
    print(f"extracted claims are saved at {output_path}")


from setuptools import setup, find_packages

setup(
   name="VeriScore",
   version="2.0.2",
   packages=find_packages(),
   install_requires=[
      'spacy',
      'openai',
      'anthropic',
      'tiktoken',
      'tqdm'
   ],
   # author="Yixio Song",
   # author_email="yixiaosong@umass.edu",
   description="Pip package for VeriScore",
   long_description=open('README.md').read(),
   long_description_content_type="text/markdown",
   # url="https://github.com/Yixiao-Song/VeriScore",
   classifiers=[
       "Programming Language :: Python :: 3",
       "License :: OSI Approved :: MIT License",
       "Operating System :: OS Independent",
   ],
   python_requires='>=3.9',
)
